unit fMain;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, cTCPServer, cHTTPServer, StdCtrls, Buttons, cWindows, ActnList,
  IniFiles, ExtCtrls, Menus, ImgList, CoolTrayIcon;

const
  WM_AFTERSHOW = WM_USER + 7223;
  AppName = 'Eternal Lines Web Server';

type
  TWebServerSettings = class
    Active       : Boolean;
    PubDir       : String;
    Port         : Integer;
    Throttle     : Boolean;
    ThrottleRate : Integer;
    LogToFile    : Boolean;
    LogFormat    : String;
    ExecEXEasCGI : Boolean;
    Minimized    : Boolean;
    FormX        : Integer;
    FormY        : Integer;

    procedure SaveToIni(Ini: TCustomIniFile; const Section: String);
    procedure LoadFromIni(Ini: TCustomIniFile; const Section: String);
  end;
  TResolveResult = (rrInvalidPath, rrNotFound, rrOK);
  TfrmMain = class(TForm)
    HTTPServer: TelHTTPServer;
    gHTTPServer: TGroupBox;
    Label1: TLabel;
    lPort: TLabel;
    Label2: TLabel;
    lUpTime: TLabel;
    Label3: TLabel;
    lRequestCount: TLabel;
    Label4: TLabel;
    Label5: TLabel;
    Label6: TLabel;
    lDataServed: TLabel;
    lConnections: TLabel;
    lLastRequest: TLabel;
    bStart: TBitBtn;
    bStop: TBitBtn;
    bSetup: TBitBtn;
    Label7: TLabel;
    lStatus: TLabel;
    tUpTime: TfndTimerHandle;
    ActionList: TActionList;
    actStart: TAction;
    actStop: TAction;
    actSetup: TAction;
    Image3: TImage;
    iServerInactive: TImage;
    iServerActive: TImage;
    TrayIcon: TCoolTrayIcon;
    mTrayIcon: TPopupMenu;
    Open1: TMenuItem;
    N1: TMenuItem;
    Start1: TMenuItem;
    Stop1: TMenuItem;
    Setup1: TMenuItem;
    N2: TMenuItem;
    Exit1: TMenuItem;
    actExit: TAction;
    actOpen: TAction;
    actAbout: TAction;
    About1: TMenuItem;
    Bevel1: TBevel;
    bExit: TBitBtn;
    bAbout: TBitBtn;
    procedure FormShow(Sender: TObject);
    procedure tUpTimeTimer(const Sender: TTimerHandle);
    procedure actSetupExecute(Sender: TObject);
    procedure actStartExecute(Sender: TObject);
    procedure actStopExecute(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure HTTPServerGetResponse(Sender: AhttpServer;
      Client: AhttpServerClient);
    procedure HTTPServerFreeResponse(Sender: AhttpServer;
      Client: AhttpServerClient);
    procedure HTTPServerClientRemoved(Sender: TTCPServerClient);
    procedure HTTPServerClientActive(Sender: TTCPServerClient);
    procedure HTTPServerPreRequest(Sender: AhttpServer;
      Client: AhttpServerClient);
    function HTTPServerCreateClient(Sender: ATCPServer): TTCPServerClient;
    procedure actExitExecute(Sender: TObject);
    procedure actOpenExecute(Sender: TObject);
    procedure TrayIconClick(Sender: TObject);
    procedure actAboutExecute(Sender: TObject);

  private
    FIni      : TIniFile;
    FSettings : TWebServerSettings;
    FLastReq  : String;
    IniLoaded : Boolean;
    FirstShow : Boolean;
    ExitApp   : Boolean;
    FLock     : TRTLCriticalSection;

    procedure Lock;
    procedure Unlock;
    procedure AfterShow(var Msg: TMessage); message WM_AFTERSHOW;
    procedure SetServerSettings;
    procedure SaveSettings;
    procedure RefreshControls;
    procedure RefreshServerStats;
    procedure RefreshUpTime;
    function  ResolvePath(const URI: String; var Path: String): TResolveResult;

  public
    property  Settings : TWebServerSettings read FSettings;
  end;

var
  frmMain: TfrmMain;

implementation

uses
  cUtils,
  cStrings,
  cDateTime,
  cWriters,
  cFileUtils,
  cInternetUtils,
  cReaders,
  cHTTP,
  cHTTPMessage,
  cWinExecute,
  fAbout,
  fSetup;

{$R *.dfm}

{ TWebServerClient                                                             }
type
  TWebServerClient = class(ThttpServerClient)
  protected
    FScriptExec  : TelWinExecute;
    FResponseStr : String;

    procedure OnCGIOutput(Sender: TWinExecute; Buf: Pointer; Size: Integer);

  public
    destructor Destroy; override;

    procedure ExecuteCGI(const RootPath, FilePath: String);
  end;

destructor TWebServerClient.Destroy;
begin
  FreeAndNil(FScriptExec);
  inherited Destroy;
end;

procedure TWebServerClient.OnCGIOutput(Sender: TWinExecute;
    Buf: Pointer; Size: Integer);
begin
  FResponseStr := FResponseStr + DupBuf(Buf^, Size);
end;

procedure TWebServerClient.ExecuteCGI(const RootPath, FilePath: String);
var S, T: String;
begin
  Response := ThttpResponse.CreateResponse(http_OK, 'OK');
  FScriptExec := TelWinExecute.Create(nil);
  FScriptExec.FilePath := FilePath;
  SetCGIEnvironment(FScriptExec.EnvStrings, RootPath, FilePath);
  FScriptExec.Options := [eoRedirectStdOut, eoNoWindow];
  FScriptExec.Mode := emBlock;
  FScriptExec.ShowMode := esHide;
  FScriptExec.DefaultDir := ExtractFilePath(FilePath);
  FScriptExec.TimeOut := 15000;
  FScriptExec.OnStdOut := OnCGIOutput;
  FResponseStr := '';
  FScriptExec.Execute;
  StrSplitAt(FResponseStr, #13#10#13#10, S, T, True, True);
  Response.ExtraHeaderLines.Text := S;
  Response.ContentAsString := T;
end;



{ TWebServerSettings                                                           }
const
  DefaultLogFormat = '$date$ $time$ $response$ $size$ $method$ $uri$';

procedure TWebServerSettings.LoadFromIni(Ini: TCustomIniFile;
  const Section: String);
begin
  with Ini do
    begin
      Active := ReadBool(Section, 'Active', True);
      PubDir := ReadString(Section, 'PubDir', 'c:\inetpub');
      Port := ReadInteger(Section, 'Port', 80);
      Throttle := ReadBool(Section, 'Throttle', False);
      ThrottleRate := ReadInteger(Section, 'ThrottleRate', 0);
      LogToFile := ReadBool(Section, 'LogToFile', True);
      LogFormat := ReadString(Section, 'LogFormat', DefaultLogFormat);
      ExecEXEAsCGI := ReadBool(Section, 'ExecEXEAsCGI', True);
      Minimized := ReadBool(Section, 'Minimized', False);
      FormX := ReadInteger(Section, 'FormX', -1);
      FormY := ReadInteger(Section, 'FormY', -1);
    end;
end;

procedure TWebServerSettings.SaveToIni(Ini: TCustomIniFile;
  const Section: String);
begin
  with Ini do
    begin
      WriteBool(Section, 'Active', Active);
      WriteString(Section, 'PubDir', PubDir);
      WriteInteger(Section, 'Port', Port);
      WriteBool(Section, 'Throttle', Throttle);
      WriteInteger(Section, 'ThrottleRate', ThrottleRate);
      WriteBool(Section, 'LogToFile', LogToFile);
      WriteString(Section, 'LogFormat', LogFormat);
      WriteBool(Section, 'ExecEXEAsCGI', ExecEXEAsCGI);
      if FormX >= 0 then
        WriteInteger(Section, 'FormX', FormX);
      if FormY >= 0 then
        WriteInteger(Section, 'FormY', FormY);
      WriteBool(Section, 'Minimized', Minimized);
    end;
end;

{ TfrmMain                                                                     }
procedure TfrmMain.FormCreate(Sender: TObject);
begin
  InitializeCriticalSection(FLock);
  FIni := TIniFile.Create('elWebServer.ini');
  FSettings := TWebServerSettings.Create;
  FSettings.LoadFromIni(FIni, 'WebServer');
  IniLoaded := True;
  if (Settings.FormX >= 0) and (Settings.FormY >= 0) then
    begin
      Position := poDesigned;
      Left := MinI(Settings.FormX, Screen.DesktopWidth - Width);
      Top := MinI(Settings.FormY, Screen.DesktopHeight - Height);
    end else
    Position := poDesktopCenter;
  FirstShow := True;
end;

procedure TfrmMain.FormDestroy(Sender: TObject);
begin
  if IniLoaded then
    begin
      Settings.FormX := MaxI(0, Left);
      Settings.FormY := MaxI(0, Top);
      SaveSettings;
    end;
  FreeAndNil(FSettings);
  FreeAndNil(FIni);
  DeleteCriticalSection(FLock);
end;

procedure TfrmMain.FormShow(Sender: TObject);
begin
  RefreshControls;
  if FirstShow then
    begin
      FirstShow := False;
      PostMessage(Handle, WM_AFTERSHOW, 0, 0);
      ShowWindow(Application.Handle, SW_HIDE);
    end;
  actOpen.Enabled := False;
end;

procedure TfrmMain.AfterShow(var Msg: TMessage);
begin
  if IniLoaded then
    begin
      if Settings.Minimized then
        begin
          actOpen.Enabled := True;
          Hide;
        end;
      if Settings.Active then
        actStart.Execute;
    end;
end;

procedure TfrmMain.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  if ExitApp then
    Action := caFree else
    begin
      Action := caNone;
      actOpen.Enabled := True;
      Settings.Minimized := True;
      Hide;
    end;
end;

procedure TfrmMain.Lock;
begin
  EnterCriticalSection(FLock);
end;

procedure TfrmMain.Unlock;
begin
  LeaveCriticalSection(FLock);
end;

procedure TfrmMain.SaveSettings;
begin
  FSettings.SaveToIni(FIni, 'WebServer');
end;

procedure TfrmMain.RefreshControls;
var R: Boolean;
begin
  R := HTTPServer.Active;
  actStart.Enabled := not R;
  actStop.Enabled := R;
  iServerActive.Visible := R;
  iServerInactive.Visible := not R;
  if R then
    begin
      lStatus.Caption := 'Active';
      lPort.Caption := HTTPServer.ListenPort;
      lConnections.Caption := IntToStr(HTTPServer.BusyClients) + ' - ' +
          IntToStr(HTTPServer.ClientCount);
    end else
    begin
      lStatus.Caption := 'Inactive';
      lPort.Caption := IntToStr(Settings.Port);
      lConnections.Caption := '-';
    end;
  RefreshServerStats;
  RefreshUpTime;
end;

procedure TfrmMain.RefreshServerStats;
var S: String;
begin
  lRequestCount.Caption := IntToStr(HTTPServer.RequestCount);
  lDataServed.Caption := StorageSize(HTTPServer.BytesServed, False);
  lLastRequest.Caption := FLastReq;
  if HTTPServer.Active then
    begin
      S := IntToStr(HTTPServer.ClientCount);
      if HTTPServer.ClientCount > 0 then
        S := S + ' (' + IntToStr(HTTPServer.BusyClients) + ' busy)';
      lConnections.Caption := S;
    end else
    lConnections.Caption := '-';
end;

procedure TfrmMain.RefreshUpTime;
begin
  if HTTPServer.Active then
    lUpTime.Caption := DateTimeAsElapsedTime(HTTPServer.UpTime) else
    lUpTime.Caption := '-';
end;

procedure TfrmMain.tUpTimeTimer(const Sender: TTimerHandle);
begin
  RefreshUpTime;
end;

procedure TfrmMain.SetServerSettings;
var R: Boolean;
begin
  if not HTTPServer.Active then
    begin
      if Settings.Port > 0 then
        HTTPServer.ListenPort := IntToStr(Settings.Port) else
        HTTPServer.ListenPort := http_DefaultPortStr;
    end;
  R := Settings.Throttle and (Settings.ThrottleRate > 0);
  HTTPServer.ThrottleClientWrite := R;
  if R then
    HTTPServer.ThrottleClientWriteRate := Settings.ThrottleRate else
    HTTPServer.ThrottleClientWriteRate := 0;
end;

procedure TfrmMain.actSetupExecute(Sender: TObject);
begin
  if ShowSetupForm then
    begin
      SetServerSettings;
      SaveSettings;
    end;
end;

procedure TfrmMain.actStartExecute(Sender: TObject);
begin
  SetServerSettings;
  HTTPServer.Active := True;
  tUpTime.TimerInterval := 1000;
  tUpTime.TimerActive := True;
  Settings.Active := True;
  RefreshControls;
  SaveSettings;
end;

procedure TfrmMain.actStopExecute(Sender: TObject);
begin
  tUpTime.TimerActive := False;
  HTTPServer.Active := False;
  Settings.Active := False;
  FLastReq := '';
  RefreshControls;
  SaveSettings;
end;

function TfrmMain.HTTPServerCreateClient(Sender: ATCPServer): TTCPServerClient;
begin
  Result := TWebServerClient.Create(Sender);
end;

const
  DefaultFiles = 4;
  DefaultFile: Array[1..DefaultFiles] of String = (
      'index.html', 'index.htm',
      'default.html', 'default.htm');

function TfrmMain.ResolvePath(const URI: String; var Path: String): TResolveResult;
var S, B : String;
    I    : Integer;
begin
  Path := '';
  Result := rrInvalidPath;
  S := Trim(URI, csAsciiCtl + [AsciiSP]);
  if S = '' then
    S := '\' else
    begin
      S := StrReplaceChar('/', '\', S);
      if PathHasDriveLetter(S) or PathIsUNCPath(S) then
        exit;
      if StrEqualNoCase(S, 'CON') or StrEqualNoCase(S, 'COM1') or
         StrEqualNoCase(S, 'COM2') or StrEqualNoCase(S, 'PRN') or
         StrEqualNoCase(S, 'AUX') then
        exit;
      if StrMatchLeft('..\', S) then
        exit;
      S := PathCanonical(S, '\');
    end;
  if Settings.PubDir = '' then
    B := '\inetpub\' else
    B := PathInclSuffix(Settings.PubDir, '\');
  S := B + StrExclPrefix(S, '\');
  if not PathIsDirectory(S) then
    begin
      if FileExists(S) then
        begin
          Path := S;
          Result := rrOK;
          exit;
        end;
      S := S + '\';
    end;
  for I := 1 to DefaultFiles do
    if FileExists(S + DefaultFile[I]) then
      begin
        Path := S + DefaultFile[I];
        Result := rrOK;
        exit;
      end;
  Path := S;
  Result := rrNotFound;
end;

procedure TfrmMain.HTTPServerGetResponse(Sender: AhttpServer;
  Client: AhttpServerClient);
var IsHead, IsGet  : Boolean;
    Method         : String;
    URI, Path, Ext : String;
begin
  Client.Synchronize(RefreshServerStats);
  Method := Client.Request.Method;
  IsHead := StrEqualNoCase(Method, http_HEAD);
  IsGet := StrEqualNoCase(Method, http_GET);
  if IsHead or IsGet then
    begin
      URI := Client.Request.URI;
      Case ResolvePath(URI, Path) of
        rrInvalidPath:
          Client.Response := httpErrorResponse(http_BadRequest,
              'Invalid path',
              httpErrorPage('Invalid path',
              'Invalid path',
              'Your client made a request with an invalid path.',
              AppName));
        rrNotFound:
          Client.Response := httpErrorResponse(http_NotFound,
              'Not found',
              httpErrorPage('Not found',
              'Not found',
              'The requested resource was not found.',
              AppName));
        rrOK:
          begin
            FLastReq := URI;
            Ext := ExtractFileExt(Path);
            if FSettings.ExecEXEasCGI and StrEqualNoCase(Ext, '.exe') then
              // execute CGI
              TWebServerClient(Client).ExecuteCGI(FSettings.PubDir, Path)
            else
              begin
                // file content
                Client.Response := ThttpResponse.CreateResponse(http_OK, 'OK');
                Client.Response.ContentType := MIMEContentTypeFromExtention(Ext);
                Client.Response.ContentAsStream := TFileReader.Create(Path);
              end;
          end;
      end;
    end else
    Client.Response := httpErrorResponse(http_MethodNotFound,
        'Method not supported',
        httpErrorPage('Method not supported',
        'Method ' + Client.Request.Method + ' not supported.',
        'Your client made a request using a method that is not recognised by ' +
        'the web server.',
        AppName));
end;

procedure TfrmMain.HTTPServerFreeResponse(Sender: AhttpServer;
  Client: AhttpServerClient);
var F: String;
    N: TDateTime;
    W: TFileWriter;
begin
  if Settings.LogToFile then
    begin
      F := Settings.LogFormat;
      if F = '' then
        F := DefaultLogFormat;
      N := Now;
      F := StrReplace('$date$', FormatDateTime('dd-mm-yyyy', N), F);
      F := StrReplace('$time$', FormatDateTime('hh:mm:ss', N), F);
      F := StrReplace('$method$', Client.Request.Method, F);
      F := StrReplace('$size$', IntToStr(Client.BytesServed), F);
      F := StrReplace('$uri$', Client.Request.URI, F);
      F := StrReplace('$referer$', Client.Request.Referer, F);
      F := StrReplace('$useragent$', Client.Request.UserAgent, F);
      if Assigned(Client.Response) then
        F := StrReplace('$response$', IntToStr(Client.Response.StatusCode), F)
      else
        F := StrReplace('$response$', '0', F);
      Lock;
      try
        W := TFileWriter.Create(GetApplicationPath + 'webserver.log');
        W.Append;
        try
          W.WriteStr(F + CRLF);
        finally
          W.Free;
        end;
      finally
        Unlock;
      end;
    end;
end;

procedure TfrmMain.HTTPServerClientActive(Sender: TTCPServerClient);
begin
  Sender.Synchronize(RefreshServerStats);
end;

procedure TfrmMain.HTTPServerClientRemoved(Sender: TTCPServerClient);
begin
  Sender.Synchronize(RefreshServerStats);
end;

procedure TfrmMain.HTTPServerPreRequest(Sender: AhttpServer;
  Client: AhttpServerClient);
begin
  Client.Synchronize(RefreshServerStats);
end;

procedure TfrmMain.actExitExecute(Sender: TObject);
begin
  ExitApp := True;
  Close;
end;

procedure TfrmMain.actOpenExecute(Sender: TObject);
begin
  Show;
  Settings.Minimized := False;
end;

procedure TfrmMain.TrayIconClick(Sender: TObject);
begin
  if actOpen.Enabled then
    actOpen.Execute;
end;

procedure TfrmMain.actAboutExecute(Sender: TObject);
begin
  ShowFrmAbout;
end;

end.

